(*
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2020)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>The Lexicographic Path Order as an Instance of WPO\<close>

text \<open>We first directly define the strict- and non-strict lexicographic path orders (LPO)
  w.r.t.\ some precedence, and then show that it is an instance of WPO.
  For this instance we use the trivial reduction pair in WPO ($\emptyset$, UNIV) and
  the status is the full one, i.e., taking parameters [0,..,n-1] for each n-ary symbol.\<close>

theory LPO
  imports
  WPO
begin

context
  fixes "pr" :: "('f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool)"
    and prl :: "'f \<times> nat \<Rightarrow> bool"
    and n :: nat
begin
fun lpo :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool \<times> bool" 
  where
  "lpo s t =
    (case s of
      Var x \<Rightarrow> (False, 
        (case t of 
          Var y \<Rightarrow> x = y 
        | Fun g ts \<Rightarrow> ts = [] \<and> prl (g, 0)))
    | Fun f ss \<Rightarrow>
        if \<exists> si \<in> set ss. snd (lpo si t) then (True, True)
        else
          (case t of
            Var _ \<Rightarrow> (False, False)
          | Fun g ts \<Rightarrow> 
            (case pr (f, length ss) (g, length ts) of (prs, prns) \<Rightarrow>            
              if prns \<and> (\<forall> tj \<in> set ts. fst (lpo s tj)) then 
                if prs then (True, True)
                else lex_ext lpo n ss ts
              else (False, False))))"

declare lpo.simps [simp del]

end


locale lpo_with_assms = precedence prc prl
  for prc :: "'f \<times> nat \<Rightarrow> 'f \<times> nat \<Rightarrow> bool \<times> bool"
    and prl :: "'f \<times> nat \<Rightarrow> bool"
    and n :: nat
begin

sublocale wpo_with_assms "{}" UNIV prc prl False "\<lambda> _. False" n full_status
  by (unfold_locales, auto simp: refl_on_def trans_def simple_arg_pos_def)

abbreviation "lpo_pr \<equiv> lpo prc prl n" 
abbreviation "lpo_s \<equiv> \<lambda> s t. fst (lpo_pr s t)"
abbreviation "lpo_ns \<equiv> \<lambda> s t. snd (lpo_pr s t)"

lemma lpo_eq_wpo: "lpo_pr s t = wpo_pr s t"
proof (induct s t rule: lpo.induct[of _ prc prl n])
  case (1 s t)
  note simps = wpo_simps[of s t] lpo.simps[of _ _ _ s t]
  show ?case 
  proof (cases s)
    case s: (Var x)
    show ?thesis unfolding simps unfolding s by (cases t, auto)
  next
    case s: (Fun f ss)
    note IH = 1[OF s]
    have id: "\<And> s. (s \<in> {}) = False" "\<And> s. (s \<in> UNIV) = True" 
      and "(\<exists>i\<in>{0..<length ss}. wpo_ns (ss ! i) t) = (\<exists>si\<in>set ss. wpo_ns si t)" 
      by (auto, force simp: set_conv_nth) 
    have IH1: "(\<exists>si \<in> set ss. wpo_ns si t) = (\<exists>si\<in>set ss. lpo_ns si t)" 
      using IH(1) by auto
    have set_index: "(\<exists>i\<in>set (\<sigma> (f, length ss)). wpo_ns (ss ! i) t) = (\<exists>si\<in>set ss. wpo_ns si t)" 
      by (auto, force simp: set_conv_nth) 
    show ?thesis unfolding simps unfolding s term.simps id if_True if_False set_index
    proof (rule sym, intro if_cong refl term.case_cong IH1, force, intro prod.case_cong refl if_cong conj_cong)
      fix g ts prs prns
      assume nstri: "\<not> (\<exists>si\<in>set ss. lpo_ns si t)" and t: "t = Fun g ts"
       and prc: "prc (f, length ss) (g, length ts) = (prs, prns)" 
      note IH = IH(2-)[OF nstri t prc[symmetric]]
      have all: "(\<forall> j \<in>{0..<length ts}. P (ts ! j)) = (\<forall> tj \<in> set ts. P tj)" for P unfolding set_conv_nth by fastforce
      show "(\<forall>j\<in>set (\<sigma> (g, length ts)). wpo_s (Fun f ss) (ts ! j)) = (\<forall>tj\<in>set ts. lpo_s (Fun f ss) tj)"
        using IH(1) by (auto simp: all[of "wpo_s (Fun f ss)"] s)
      assume "prns \<and> (\<forall>tj\<in>set ts. lpo_s (Fun f ss) tj)"  "\<not> prs"
      note IH = IH(2)[OF this[folded s]]
      show "lex_ext wpo_pr n (map ((!) ss) (\<sigma> (f, length ss))) (map ((!) ts) (\<sigma> (g, length ts))) = lex_ext lpo_pr n ss ts" 
        using IH by (simp add: map_nth, auto simp: lex_ext_iff)
    qed
  qed
qed

abbreviation "LPO_S \<equiv> {(s,t). lpo_s s t}"
abbreviation "LPO_NS \<equiv> {(s,t). lpo_ns s t}"

theorem LPO_SN_order_pair: "SN_order_pair LPO_S LPO_NS"
  unfolding lpo_eq_wpo by (rule WPO_SN_order_pair)

theorem LPO_S_subst: "(s,t) \<in> LPO_S \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> LPO_S" for \<sigma> :: "('f,'a)subst" 
  using WPO_S_subst unfolding lpo_eq_wpo .

theorem LPO_NS_subst: "(s,t) \<in> LPO_NS \<Longrightarrow> (s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> LPO_NS" for \<sigma> :: "('f,'a)subst"
  using WPO_NS_subst unfolding lpo_eq_wpo .

theorem LPO_NS_ctxt: "(s,t) \<in> LPO_NS \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> LPO_NS" 
  using WPO_NS_ctxt unfolding lpo_eq_wpo .

theorem LPO_S_ctxt: "(s,t) \<in> LPO_S \<Longrightarrow> (Fun f (bef @ s # aft), Fun f (bef @ t # aft)) \<in> LPO_S" 
  using WPO_S_ctxt unfolding lpo_eq_wpo by auto

theorem LPO_S_subset_LPO_NS: "LPO_S \<subseteq> LPO_NS" 
  using WPO_S_subset_WPO_NS unfolding lpo_eq_wpo .

theorem supt_subset_LPO_S: "{\<rhd>} \<subseteq> LPO_S" 
  using supt_subset_WPO_S unfolding lpo_eq_wpo by auto

theorem supteq_subset_LPO_NS: "{\<unrhd>} \<subseteq> LPO_NS" 
  using supteq_subset_WPO_NS unfolding lpo_eq_wpo by auto

end

end
